from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    NamedTuple,
    Optional,
    Type,
    TypeVar,
    Union,
)

from .interface import Container
from .scope_enum import ScopeEnum
from .static_container_builder import StaticContainerBuilder


TARGET_T = TypeVar("TARGET_T")
PROVIDER_T = TypeVar("PROVIDER_T")
FACTORY_T = Callable[..., PROVIDER_T]


class ProviderBinding(NamedTuple):
    """Represents a binding for providing instances through dependency injection."""

    implementation: Type[PROVIDER_T]
    annotation: Optional[Type[PROVIDER_T]] = None
    scope: Optional[Union[str, ScopeEnum]] = None
    on_activate: Optional[Callable[[PROVIDER_T], PROVIDER_T]] = None


class ConstantBinding(NamedTuple):
    """Represents a binding for providing constant values through dependency injection."""

    value: PROVIDER_T
    annotation: Type[PROVIDER_T]


class FactoryBinding(NamedTuple):
    """Represents a binding for providing instances created by factory functions."""

    factory: Callable[[Container], FACTORY_T]
    annotation: FACTORY_T


def _add_provider_to_map(binding_map: Dict, binding: ProviderBinding):

    annotation = binding.annotation or binding.implementation
    scope = binding.scope or ScopeEnum.REQUESTED

    if isinstance(scope, str):
        try:
            scope = {
                "TRANSIENT": ScopeEnum.TRANSIENT,
                "REQUESTED": ScopeEnum.REQUESTED,
                "SINGLETON": ScopeEnum.SINGLETON,
                "transient": ScopeEnum.TRANSIENT,
                "requested": ScopeEnum.REQUESTED,
                "singleton": ScopeEnum.SINGLETON,
                "t": ScopeEnum.TRANSIENT,
                "r": ScopeEnum.REQUESTED,
                "s": ScopeEnum.SINGLETON,
                "T": ScopeEnum.TRANSIENT,
                "R": ScopeEnum.REQUESTED,
                "S": ScopeEnum.SINGLETON,
            }[scope]

        except KeyError as ex:
            raise ValueError(f'Unknown scope "{scope}"') from ex

    binding_map[annotation] = ProviderBinding(
        implementation=binding.implementation,
        annotation=annotation,
        scope=scope,
        on_activate=binding.on_activate
    )


def _add_binding_to_map(
    binding_map: Dict,
    binding: Union[ProviderBinding, ConstantBinding, FactoryBinding]
):
    if isinstance(binding, ProviderBinding):
        _add_provider_to_map(binding_map, binding)
    else:
        binding_map[binding.annotation] = binding


class BuilderBase(Generic[TARGET_T]):
    """Base class for building instances with dependency injection.

    Example Usage:

        class DuckInterface(ABC):
            ...

        class DuckBuilder(Builder[DuckInterface]):
            def __init__(self) -> None:
                super().__init__(DuckInterface)

        class Mallard(DuckInterface):
            ...

        duck = (
            DuckBuilder()
            .using_provider(Mallard, DuckInterface)
            .build()
        )

        assert isinstance(duck, Mallard)
    """

    def __init__(
        self,
        target_t: Type[TARGET_T],
        providers: Optional[List[ProviderBinding]] = None,
        constants: Optional[List[ConstantBinding]] = None,
        factories: Optional[List[FactoryBinding]] = None,
    ):
        """
        Initialize the builder with optional providers, constants, and factories.

        Arguments:
            providers: An optional list of provider bindings.
            constants: An optional list of constant bindings.
            factories: An optional list of factory bindings.
        """
        self._target_t = target_t
        self._providers: Dict[Type, ProviderBinding] = {}
        self._constants: Dict[Type, ConstantBinding] = {}
        self._factories: Dict[Type, FactoryBinding] = {}

        for provider in providers or []:
            _add_binding_to_map(self._providers, provider)

        for constant in constants or []:
            _add_binding_to_map(self._constants, constant)

        for factory in factories or []:
            _add_binding_to_map(self._factories, factory)

    def using_provider(
        self,
        implementation: Type[PROVIDER_T],
        annotation: Optional[Type[PROVIDER_T]] = None,
        scope: Optional[Union[str, ScopeEnum]] = None,
        on_activate: Optional[Callable[[PROVIDER_T], PROVIDER_T]] = None,
    ) -> "BuilderABC(Generic[BUILDER_T])":
        """
        Register a provider for dependency injection.

        Arguments:
            implementation: The implementation class.
            annotation: The optional annotion target.
            scope: The optional scope of the provider. Defaults to Requested.
            on_activate: An optional activation callback.

        Returns:
            BuilderABC: The builder instance.
        """
        _add_provider_to_map(
            self._providers,
            ProviderBinding(
                implementation=implementation,
                annotation=annotation,
                scope=scope,
                on_activate=on_activate,
            )
        )
        return self

    def using_constant(
        self,
        value: PROVIDER_T,
        annotation: Type[PROVIDER_T],
    ) -> "BuilderABC(Generic[BUILDER_T])":
        """
        Register a constant for dependency injection.

        Arguments:
            value: The constant value.
            annotation: The annotion target.

        Returns:
            BuilderABC: The builder instance.
        """
        _add_binding_to_map(
            self._constants,
            ConstantBinding(
                value=value,
                annotation=annotation,
            )
        )
        return self

    def using_factory(
        self,
        factory: Callable[[Container], FACTORY_T],
        annotation: FACTORY_T,
    ) -> "BuilderABC(Generic[BUILDER_T])":
        """
        Register a factory function for dependency injection.

        Arguments:
            factory: The factory function.
            annotation: The annotion target.

        Returns:
            BuilderABC: The builder instance.
        """
        _add_binding_to_map(
            self._factories,
            FactoryBinding(
                factory=factory,
                annotation=annotation,
            )
        )
        return self

    def build(self) -> TARGET_T:
        """
        Build an instance of the specified target type using the registered dependencies.

        Returns:
            TARGET_T: The built instance.
        """

        container_builder = StaticContainerBuilder()

        for provider in self._providers.values():
            container_builder.bind(
                annotation=provider.annotation,
                implementation=provider.implementation,
                scope=provider.scope,
                on_activate=provider.on_activate,
            )

        for constant in self._constants.values():
            container_builder.bind_constant(
                annotation=constant.annotation,
                value=constant.value,
            )

        for factory in self._factories.values():
            container_builder.bind_factory(
                annotation=factory.annotation,
                factory=factory.factory,
            )

        container = container_builder.build()
        return container.get(self._target_t)




from collections import deque

from .scope_enum import ScopeEnum
from .bound_member import BoundMember
from .scope_container import PersistentScope, ScopeContainer
from .interface import Container


class StaticContainer(Container):

    def __init__(self, bound_members):
        self._singletons = PersistentScope()
        self._bound_members = bound_members

    def _build_scope(self, requested_member: BoundMember):

        # Build a scope using a post-order traversal of the
        # dependency tree.  This will guarantee the scope has
        # all dependencies for each object it is given to build.

        scope = ScopeContainer(self._singletons)
        stack = deque()
        stack.append((requested_member, 0))
        while len(stack) > 0:
            m, s = stack.pop()
            if m.scope != ScopeEnum.TRANSIENT and scope.has(m):
                continue
            elif s == 0:
                stack.append((m, 1))
                [stack.append((v, 0)) for v in m]
            else:
                scope.add(m)
        return scope

    def get(self, annotation):
        member = self._bound_members[annotation]
        scope = self._build_scope(member)
        return scope.get_instance_of(member)



from abc import ABC, abstractmethod
from typing import Callable, Dict, List

from .scope_enum import ScopeEnum
from .bound_member import BoundMember


class Scope(ABC):
    """An IOC Instance Scope"""

    @abstractmethod
    def __contains__(self, annotation: any) -> bool:
        """Check if the given annotation exists in this scope.

        Arguments:
         annotation: The str or type var used to identify an instance.

        Returns:
         True, if the annotation exists. Else, False.
        """
        raise NotImplementedError()

    @abstractmethod
    def add(self, annotation: any, instance: object) -> None:
        """Add a instance to this scope under the given annotation

        Arguments:
         annotation: The str or type var used to identify an instance.
         instance: Any value to store as the instance associated with the annotation.
        """
        raise NotImplementedError()

    @abstractmethod
    def use(self, annotation: any) -> object:
        """Get the instance associated with the given annotation.

        Arguments:
         annotation: The str or type var used to identify an instance.

        Raises:
         KeyError, if the the given annotation does not exist in this scope.
        """
        raise NotImplementedError()


class Container(ABC):
    """An IOC Container"""

    @abstractmethod
    def get(self, annotation: any) -> object:
        """Retrieve an instance from the container

        The instance will be produced according to it's scope. If the instance
        is new, it's dependency chain will also be created according to their scope.

        Arguments:
         annotation: The hint used to locate the member

        Returns:
         An object instance, constant, or function
        """
        raise NotImplementedError()


class ContainerBuilder(ABC):
    """Bind classes, values, functions, and factories to a container."""

    @abstractmethod
    def bind(
        self,
        annotation: any,
        implementation: Callable,
        scope: ScopeEnum = ScopeEnum.TRANSIENT,
        on_activate: Callable[[any], any] = None,
    ):
        """Bind a class.

        Bind any callable type to an annotation. Dependencies will be
        injected into this object as needed when created.

        Scoping can be set to control reuse.

        Arguments:
          annotation:     The hint used to inject an instance of implementation

          implementation: A callable type who's result will be stored return and
                          stored according to the scope

          scope:          Identifies how the object should be cached. Options are
                          Transient, Requested, Singleton
                          Default: Transient.

          on_activate:    An optional function that will be called with the
                          constructed implementation before it is used as a dep
                          or given as the return in container.get()
                          Default: None.

        Scopes:
            Transient scopes and not cached.
            Requested scopes are cached during the current execution of a container.get call.
            Singleton scopes are only instanced once and cached for the lifetime of the container.

        Example:

            class Duck:
                def quack(self):
                    print("quack")

            ioc_builder.bind(
                annotation="duck",
                implementation=Duck)
        """
        raise NotImplementedError()

    @abstractmethod
    def bind_constant(self, annotation: any, value: object):
        """Bind a constant value

        This allows you to bind any object to an annotation in a singleton scope.

        Arguments:
          annotation: The hint used to inject the constant
          value: Any value. Object, function, type, anything.

        Example:

            ioc_builder.bind_constant(
                annotation="my_constant",
                value="Hello, world!")

        """
        raise NotImplementedError()

    @abstractmethod
    def bind_factory(self, annotation: any, factory: Callable[[Container], any]):
        """Bind a higher order function

        This approach allows you to control the creation of objects and gives you access
        to the container. This lets you make runtime decision about how to create an instance.

        Arguments:
          annotation: The hint used to inject the factory
          factory:    A higher order function that accepts the StackContainer as an
                      arugment.

        Example:
            def my_factory_wrapper(ctx: StaticContainer)

                def my_factory(foo):
                    bar = ctx.get("bar")
                    bar.baz(foo)
                    return bar

                return my_factory

            ioc_builder.bind_factory(
                annotation="my_factory",
                factory=my_function)
        """
        raise NotImplementedError()

    @abstractmethod
    def build(self) -> Container:
        """Compute dependency graph and return the container

        This call will roll over all the objects and compute the dependants of each
        member. The container itself is also added to the graph and can thus be
        injected using it's Type as the annotation.

        Example:
            ioc_builder = StaticContainerBuilder()
            ioc = ioc_builder.build()
            container = ioc.get(StaticContainer)
            container == ioc ## True
        """
        raise NotImplementedError()


class CycleTest(ABC):
    """Test a graph of BoundMembers for circular dependencies."""
    @abstractmethod
    def find_cycle(self, bound_members: Dict[any, BoundMember]) -> List[BoundMember]:
        """Check graph of BoundMembers for cycles.

        Arguments:
          bound_members: The graph to test.
        """
        raise NotImplementedError()

class BoundMemberFactory(ABC):
    @abstractmethod
    def build(
        self,
        annotation: any,
        implementation: Callable,
        scope: ScopeEnum
    ) -> BoundMember:
        """Create a new bound member.

        Arguments:
          annotation: The hint associated with this member
          implementation: The implementation used for this member.
          scope: The scope used to construct implementations for this member.
        """
        raise NotImplementedError()



from .bound_member import BoundMember
from .interface import Scope
from .scope_enum import ScopeEnum

class PersistentScope(Scope):

    def __init__(self):
        self._cache = {}

    def __contains__(self, annotation) -> bool:
        return annotation in self._cache.keys()

    def add(self, annotation, instance) -> None:
        self._cache[annotation] = instance

    def use(self, annotation) -> object:
        return self._cache[annotation]


class TransientScope(Scope):

    def __init__(self):
        self._cache = {}

    def __contains__(self, annotation) -> bool:
        return annotation in self._cache.keys()

    def add(self, annotation, instance) -> None:
        if annotation in self:
            self._cache[annotation].append(instance)
        else:
            self._cache[annotation] = [instance]

    def use(self, annotation) -> object:
        inst = self._cache[annotation].pop()
        if len(self._cache[annotation]) == 0:
            del self._cache[annotation]
        return inst


class ScopeContainer:

    def __init__(self, singleton: Scope):
        self._scopes = {
            ScopeEnum.SINGLETON: singleton,
            ScopeEnum.REQUESTED: PersistentScope(),
            ScopeEnum.TRANSIENT: TransientScope()}

    def _create_instance(self, member: BoundMember) -> object:
        args = list()
        for dep in member:
            args.append(self.get_instance_of(dep))
        return member.on_activate(member.implementation(*args))

    def _get_scope(self, member: BoundMember):
        return self._scopes[member.scope]

    def has(self, member: BoundMember) -> bool:
        return member.annotation in self._get_scope(member)

    def add(self, member: BoundMember):
        self._get_scope(member).add(member.annotation, self._create_instance(member))

    def get_instance_of(self, member: BoundMember) -> object:
        return self._get_scope(member).use(member.annotation)



from inspect import signature, isclass
from types import FunctionType, MethodType
from typing import get_type_hints, Callable, Any

from .bound_member import BoundMember
from .scope_enum import ScopeEnum
from .interface import BoundMemberFactory


class DefaultBoundMemberFactory(BoundMemberFactory):

    def build(
        self,
        annotation: any,
        implementation: Callable,
        scope: ScopeEnum,
        on_activate: Callable[[Any], Any] = None,
    ) -> BoundMember:

        if isclass(implementation):
            params = get_type_hints(implementation.__init__)
        elif type(implementation) == FunctionType:
            params = get_type_hints(implementation)
        elif type(implementation) == MethodType:
            params = get_type_hints(implementation)
        elif hasattr(implementation, '__call__'):
            params = get_type_hints(implementation.__call__)
        else:
            params = get_type_hints(implementation)

        return BoundMember(
            annotation=annotation,
            implementation=implementation,
            scope=scope,
            parameters=[p for k, p in params.items() if k != 'return'],
            on_activate=on_activate
        )



from setuptools import setup


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="pyioc3",
    version="1.4.0",
    description="Python IOC Container",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ian Laird",
    author_email="irlaird@gmail.com",
    url="https://github.com/en0/pyioc",
    packages=["pyioc3"],
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
